@@ -29,11 +29,11 @@
 #include <stdlib.h>
 #include <stdarg.h>
 #include <unistd.h>
-#include <time.h>
 #include <errno.h>
 #include <sys/types.h>
 #include <assert.h>
 #include <limits.h>
+#include <regex.h>
 
 #include <sqlite3.h>
 #include <libprelude/prelude.h>
