@@ -62,14 +62,20 @@
 #include "preludedb.h"
 
 
+typedef enum {
+	PRELUDEDB_SQL_STATUS_CONNECTED    = 0x01,
+	PRELUDEDB_SQL_STATUS_TRANSACTION  = 0x02
+} preludedb_sql_status_t;
 
-#define assert_connected(sql)                                   \
-        if ( sql->status < PRELUDEDB_SQL_STATUS_CONNECTED ) {   \
-                int __ret;                                      \
-                                                                \
-                __ret = preludedb_sql_connect(sql);             \
-                if ( __ret < 0 )                                \
-                        return __ret;                           \
+
+
+#define assert_connected(sql)                                     \
+        if ( ! (sql->status & PRELUDEDB_SQL_STATUS_CONNECTED) ) { \
+                int __ret;                                        \
+                                                                  \
+                __ret = preludedb_sql_connect(sql);               \
+                if ( __ret < 0 )                                  \
+                        return __ret;                             \
         }
 
 
@@ -117,21 +123,12 @@ extern prelude_list_t _sql_plugin_list;
 
 
 
-static void preludedb_sql_disconnect(preludedb_sql_t *sql)
-{
-        _preludedb_plugin_sql_close(sql->plugin, sql->session);
-
-        sql->status = PRELUDEDB_SQL_STATUS_DISCONNECTED;
-        sql->session = NULL;
-}
-
-
-
-
 static inline void update_sql_from_errno(preludedb_sql_t *sql, preludedb_error_t error)
 {
-        if ( preludedb_error_check(error, PRELUDEDB_ERROR_CONNECTION) )
-                preludedb_sql_disconnect(sql);
+        if ( preludedb_error_check(error, PRELUDEDB_ERROR_CONNECTION) ) {
+                _preludedb_plugin_sql_close(sql->plugin, sql->session);                
+                sql->status &= ~PRELUDEDB_SQL_STATUS_CONNECTED;
+        }
 }
 
 
@@ -190,7 +187,7 @@ int preludedb_sql_new(preludedb_sql_t **new, const char *type, preludedb_sql_set
  */
 void preludedb_sql_destroy(preludedb_sql_t *sql)
 {
-        if ( sql->status >= PRELUDEDB_SQL_STATUS_CONNECTED )
+        if ( sql->status & PRELUDEDB_SQL_STATUS_CONNECTED )
                 _preludedb_plugin_sql_close(sql->plugin, sql->session);
 
         if ( sql->logfile )
@@ -351,11 +348,9 @@ int preludedb_sql_query(preludedb_sql_t *sql, const char *query, preludedb_sql_t
 int preludedb_sql_query_sprintf(preludedb_sql_t *sql, preludedb_sql_table_t **table,
                                 const char *format, ...)
 {
+        int ret;
         va_list ap;
         prelude_string_t *query;
-        int ret;
-
-        assert_connected(sql);
 
         ret = prelude_string_new(&query);
         if ( ret < 0 )
@@ -364,6 +359,7 @@ int preludedb_sql_query_sprintf(preludedb_sql_t *sql, preludedb_sql_table_t **ta
         va_start(ap, format);
         ret = prelude_string_vprintf(query, format, ap);
         va_end(ap);
+        
         if ( ret < 0 )
                 goto error;
 
@@ -392,11 +388,9 @@ int preludedb_sql_query_sprintf(preludedb_sql_t *sql, preludedb_sql_table_t **ta
 int preludedb_sql_insert(preludedb_sql_t *sql, const char *table, const char *fields,
                          const char *format, ...)
 {
+        int ret;
         va_list ap;
         prelude_string_t *query;
-        int ret;
-
-        assert_connected(sql);
 
         ret = prelude_string_new(&query);
         if ( ret < 0 )
@@ -417,8 +411,6 @@ int preludedb_sql_insert(preludedb_sql_t *sql, const char *table, const char *fi
                 goto error;
 
         ret = preludedb_sql_query(sql, prelude_string_get_string(query), NULL);
-        if ( ret < 0 )
-                update_sql_from_errno(sql, ret);
 
  error:
         prelude_string_destroy(query);
@@ -450,18 +442,14 @@ int _preludedb_sql_transaction_start(preludedb_sql_t *sql)
 {
         int ret;
                 
-        if ( sql->status == PRELUDEDB_SQL_STATUS_TRANSACTION )                
+        if ( sql->status & PRELUDEDB_SQL_STATUS_TRANSACTION )                
                 return preludedb_error(PRELUDEDB_ERROR_ALREADY_IN_TRANSACTION);
         
-        assert_connected(sql);
-
         ret = preludedb_sql_query(sql, "BEGIN", NULL);
-        if ( ret < 0 ) {
-                update_sql_from_errno(sql, ret);
+        if ( ret < 0 )
                 return ret;
-        }
         
-        sql->status = PRELUDEDB_SQL_STATUS_TRANSACTION;
+        sql->status |= PRELUDEDB_SQL_STATUS_TRANSACTION;
 
         return 0;
 }
@@ -490,15 +478,11 @@ int _preludedb_sql_transaction_end(preludedb_sql_t *sql)
 {
         int ret;
 
-        if ( sql->status != PRELUDEDB_SQL_STATUS_TRANSACTION )                
+        if ( ! (sql->status & PRELUDEDB_SQL_STATUS_TRANSACTION) )                
                 return preludedb_error(PRELUDEDB_ERROR_NOT_IN_TRANSACTION);
         
         ret = preludedb_sql_query(sql, "COMMIT", NULL);
-
-        sql->status = PRELUDEDB_SQL_STATUS_CONNECTED;
-
-        if ( ret < 0 )
-                update_sql_from_errno(sql, ret);
+        sql->status &= ~PRELUDEDB_SQL_STATUS_TRANSACTION;
 
         return ret;
 }
@@ -529,26 +513,32 @@ int _preludedb_sql_transaction_abort(preludedb_sql_t *sql)
         int ret;
         char *original_error = NULL;
         
-        if ( sql->status != PRELUDEDB_SQL_STATUS_TRANSACTION )
+        if ( ! (sql->status & PRELUDEDB_SQL_STATUS_TRANSACTION) )
                 return preludedb_error(PRELUDEDB_ERROR_NOT_IN_TRANSACTION);
-        
+
         if ( _prelude_thread_get_error() ) 
                 original_error = strdup(_prelude_thread_get_error());
-        
-        ret = preludedb_sql_query(sql, "ROLLBACK", NULL);
-        sql->status = PRELUDEDB_SQL_STATUS_CONNECTED;
 
-        if ( ret < 0 ) {
-                update_sql_from_errno(sql, ret);
+        sql->status &= ~PRELUDEDB_SQL_STATUS_TRANSACTION;
 
-                if ( ! original_error )
-                        ret = preludedb_error(PRELUDEDB_ERROR_QUERY);
-                else {
-                        ret = preludedb_error_verbose(PRELUDEDB_ERROR_QUERY, "%s", original_error);
-                        free(original_error);
-                }
+        if ( original_error && ! (sql->status & PRELUDEDB_SQL_STATUS_CONNECTED) ) {
+                ret = preludedb_error_verbose(PRELUDEDB_ERROR_QUERY, "%s. No ROLLBACK possible due to connection closure", 
+                                              original_error);
+                free(original_error);
+                return ret;
         }
         
+        ret = preludedb_sql_query(sql, "ROLLBACK", NULL);
+        if ( ret < 0 ) {                
+                if ( original_error )
+                        ret = preludedb_error_verbose(PRELUDEDB_ERROR_QUERY, "%s.\nROLLBACK failed: %s", original_error, preludedb_strerror(ret));
+                else
+                        ret = preludedb_error_verbose(PRELUDEDB_ERROR_QUERY, "ROLLBACK failed: %s", preludedb_strerror(ret));
+        }
+
+        if ( original_error )
+                free(original_error);
+                        
         return ret;
 }
 
@@ -574,7 +564,7 @@ int preludedb_sql_transaction_abort(preludedb_sql_t *sql)
 
 
 /**
- * preludedb_sql_escape:
+ * preludedb_sql_escape_fast:
  * @sql: Pointer to a sql object.
  * @input: Buffer to escape
  * @input_size: Buffer size.
@@ -604,8 +594,6 @@ int preludedb_sql_escape_fast(preludedb_sql_t *sql, const char *input, size_t in
  */
 int preludedb_sql_escape(preludedb_sql_t *sql, const char *input, char **output) 
 {
-        assert_connected(sql);
-
         if ( ! input ) {
                 *output = strdup("NULL");
                 return *output ? 0 : preludedb_error_from_errno(errno);
@@ -801,8 +789,8 @@ unsigned int preludedb_sql_table_get_row_count(preludedb_sql_table_t *table)
  */
 int preludedb_sql_table_fetch_row(preludedb_sql_table_t *table, preludedb_sql_row_t **row)
 {
-        void *res;
         int ret;
+        void *res;
 
         ret = _preludedb_plugin_sql_fetch_row(table->sql->plugin, table->sql->session, table->res, &res);
         if ( ret < 0 ) {
